#define BPF_NO_GLOBAL_DATA
#define __TARGET_ARCH_x86
#include <linux/bpf.h>
#include <bpf/bpf_helpers.h>
#include <bpf/bpf_tracing.h>
#include <bpf/bpf_core_read.h>
#include <linux/if_ether.h>
#include <linux/udp.h>
#include <linux/net.h>
#include <linux/socket.h>
#include <sys/socket.h>
#include <linux/net.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include "common.h"


/*
 * Helper:
 * Issue: invalid indirect read from stack R2 off
 *   Fix: check if all variables is initialised
 * Issue: R1 invalid mem access 'inv'
 *   Fix: the value can be NULL
 */

struct {
    __uint(type, BPF_MAP_TYPE_RINGBUF);
    __uint(max_entries, 256 * 1024 /* 256kb */);
} m_data SEC(".maps");

/*
 * https://datatracker.ietf.org/doc/html/rfc1035
 */
static int dnsquery(struct __sk_buff *skb, struct ethhdr eth, struct iphdr ip, struct udphdr udp, int dport, int sport, int tlen){
    struct event s_event = {0};
    struct dnshdr dns = {0};
    size_t dlen = 0;
    size_t qlen = 0; // Question len
    s_event.dport = dport;
    s_event.sport = sport;
    bpf_printk("port: %d %d", s_event.dport, s_event.sport);
    //if (data + sizeof(struct ethhdr) + sizeof(struct iphdr) + sizeof(struct udphdr) + sizeof(struct dnshdr) > data_end)
    //    return 0;

    //dns = (struct dnshdr *)(data + sizeof(struct ethhdr) + sizeof(struct iphdr) + sizeof(struct udphdr));
    //tlen += sizeof(struct dnshdr);
    bpf_printk("udp len: %d", ntohs(udp.len));
    bpf_skb_load_bytes(skb, sizeof(struct ethhdr) + sizeof(struct iphdr) + sizeof(struct udphdr), &dns, sizeof(struct dnshdr));

    if (ntohs(dns.nbQuestions) == 0)
        return 0;

    bpf_printk("tid: %x", ntohs(dns.transactionID)); // Use as key map
    bpf_printk("nb question: %d", ntohs(dns.nbQuestions));

    struct dns_query dquery;
    bpf_skb_load_bytes(skb, sizeof(struct ethhdr) + sizeof(struct iphdr) + sizeof(struct udphdr) + sizeof(struct dnshdr), &dquery, sizeof(struct dns_query));
    bpf_printk("%s", dquery.name);
    bpf_printk("class: %d", ntohs(dquery.class));
    bpf_printk("type: %d", ntohs(dquery.type));
    // bpf_printk("size: %d %d %d", tlen, skb->len, (skb->len - tlen));
    dlen = (skb->len - tlen);
    bpf_printk("DNS packet len: %d", dlen);
    qlen = dlen - sizeof(struct dnshdr);
    bpf_printk("size: %d %d", sizeof(struct dnshdr), qlen);
    char buf[128] = {0};


    bpf_skb_load_bytes(skb, sizeof(struct ethhdr) + sizeof(struct iphdr) + sizeof(struct udphdr) + sizeof(struct dnshdr), &buf, 41);
    int index = 0;
    char c = 0;
    char t[15];
    while (index != 15){
        c = buf[index];
        if (c == '\0')
            break;
        t[index] = c;
        if (c == 0x3)
            t[index] = '.';
        index++;
    }
    bpf_printk("%s", t);
    // https://docs.cilium.io/en/stable/reference-guides/bpf/progtypes/
    if (bpf_ringbuf_output(&m_data, &s_event, sizeof(struct event), 0) == 0)
        return 0;
    //s_event = bpf_ringbuf_reserve(&m_data, sizeof(*s_event), 0); 
    //if (!s_event)
    //    return 0;
    //bpf_ringbuf_discard(s_event, 0);

    //bpf_ringbuf_submit(s_event, 0);
    return 0;
}

SEC("socket")
int detect_dns(struct __sk_buff *skb) {
    //void *data = (void *)(long)skb->data;
    //void *data_end = (void *)(long)skb->data_end;
    //struct ethhdr *eth = data;
    struct ethhdr eth = {0};
    struct iphdr ip = {0};
    struct udphdr udp = {0};
    unsigned long long h_proto, p;
    unsigned long long dport;
    unsigned long long sport;
    size_t tlen = 0;

    //if (data + sizeof(struct ethhdr) + sizeof(struct iphdr) + sizeof(struct udphdr) > data_end)
    //    return 0;

    //bpf_skb_load_bytes(skb, 12, &p, 2);
    bpf_skb_load_bytes(skb, 0, &eth, sizeof(struct ethhdr));
    p = eth.h_proto;
    tlen += sizeof(struct ethhdr);
    if (ntohs(p) != ETH_P_IP)
        return 0;

    tlen += sizeof(struct iphdr);
    // bpf_printk("ip: %d",ntohs(p));

    //ip = (struct iphdr*)(data + sizeof(struct ethhdr));
    bpf_skb_load_bytes(skb, sizeof(struct ethhdr), &ip, sizeof(struct iphdr));

    //bpf_skb_load_bytes(data, sizeof(struct ethhdr), ip, sizeof(struct ip));
    h_proto = ip.protocol;
    // If not UDP packet
    if (h_proto != 17) 
        return 0;

    bpf_printk("proto: %d", h_proto);
    tlen += sizeof(struct udphdr);
    //udp = (struct udphdr*)(data + sizeof(struct ethhdr) + sizeof(struct iphdr));

    bpf_skb_load_bytes(skb, sizeof(struct ethhdr) + sizeof(struct iphdr), &udp, sizeof(struct udphdr));

    // Check if DNS port
    dport = ntohs(udp.dest);
    sport = ntohs(udp.source);
    
    //if (dport != 53)
    //    return 0;
     
    if (dport == 53)
        dnsquery(skb, eth, ip, udp, dport, sport, tlen);
    else if(sport == 53)
        bpf_printk("Response");

    return 0;
}
/*SEC("xdp")
int detect_dns(struct xdp_md *ctx){
    void *data_end = (void *)(long)ctx->data_end;
    void *data = (void *)(long)ctx->data;
    struct ethhdr *eth = data;
    struct iphdr *ip;
    struct udphdr *udp;
    __u16 h_proto;
    __u16 dport;

    if (data + sizeof(struct ethhdr) + sizeof(struct iphdr) + sizeof(struct udphdr) > data_end)
        return XDP_DROP;

    ip = (struct iphdr*)(data + sizeof(struct ethhdr));
    udp = (struct udphdr*)(data + sizeof(struct ethhdr) + sizeof(struct iphdr));

    h_proto = ip->protocol;
    // If not UDP packet
    if (h_proto != 17) 
        return XDP_PASS;

    // Check if DNS port
    dport = udp->dest;
    bpf_printk("Dport: %d", (dport));

    return XDP_PASS;
}*/
char LICENSE[] SEC("license") = "GPL";
